#!/usr/bin/python3

import math
import numpy
import random
import transformations as t3d

class CityGenerator:

    '''
       Constructor
    '''
    def __init__(self, enable_blinders = False):
        self.enable_blinders = enable_blinders          # flag to toggle blinders between house gaps

        ## parameters as used in the user study described in the paper ##
        self.path_width = 10.0                          # width of streets in meters
        self.min_extension = 20.0                       # minimum house length to be applied after l3
        self.max_additional_extension = 100.0           # maximum additional house length after l3 in
                                                        # addition to self.min_extension
        self.min_baseline_distance = 15.0               # minimum length in meters of baseline route
        self.max_baseline_distance = 20.0               # maximum length in meters of baseline route
        self.decorator_probability = 0.2                # probability that a decorator object is placed
                                                        # in front of a house
        self.house_scale_vector = [7.5, 15.0, 7.5]      # unit cube scaling to resemble house shapes
        
        ## helper variables ##
        scl_mat_x = t3d.scale_matrix(self.house_scale_vector[0], origin=None, direction=[1, 0, 0])
        scl_mat_y = t3d.scale_matrix(self.house_scale_vector[1], origin=None, direction=[0, 1, 0])
        scl_mat_z = t3d.scale_matrix(self.house_scale_vector[2], origin=None, direction=[0, 0, 1])
        self.house_scale_matrix = scl_mat_x * scl_mat_y * scl_mat_z

        self.house_matrices = []
        self.marker_matrices = []
        self.decorator_matrices = []


    '''
       Creates the lists of house and marker transformation matrices.
    '''
    def generate_transformation_matrices(self, l1, l2, l3, turn_mode, 
                                         house_output_file, decorator_output_file, marker_output_file):
        print("Route parameters: " + str((l1,l2,l3)) + " " + str(turn_mode))
        self.create_object_matrices(l1, l2, l3, turn_mode)
        self.create_marker_matrices(l1, l2, l3, turn_mode)
        self.save_matrices_to_csv(house_output_file, decorator_output_file, marker_output_file)
        print("Success")


    '''
       Saves the lists of previously created house and marker transformation matrices.
    '''
    def save_matrices_to_csv(self, house_output_file, decorator_output_file, marker_output_file):
        print("Saving house matrices...")
        file_handle = open(house_output_file, 'w')
        for matrix in self.house_matrices:
            mat_list = matrix.tolist()
            mat_list = [[round(val, 3) for val in row] for row in mat_list]
            file_handle.write(str(mat_list) + '\n')
        file_handle.close()

        print("Saving decorator matrices...")
        file_handle = open(decorator_output_file, 'w')
        for matrix in self.decorator_matrices:
            mat_list = matrix.tolist()
            mat_list = [[round(val, 3) for val in row] for row in mat_list]
            file_handle.write(str(mat_list) + '\n')
        file_handle.close()

        print("Saving route markers matrices...")
        file_handle = open(marker_output_file, 'w')
        for matrix in self.marker_matrices:
            mat_list = matrix.tolist()
            mat_list = [[round(val, 3) for val in row] for row in mat_list]
            file_handle.write(str(mat_list) + '\n')
        file_handle.close()


    '''
       Returns the list of previously created house transformation matrices.
    '''
    def get_house_matrices(self):
        if self.house_matrices is []:
            print("Markers need to be created first! Call 'create_house_matrices()' first")
        return self.house_matrices


    '''
       Returns the list of previously created decorator transformation matrices.
    '''
    def get_decorator_matrices(self):
        if self.decorator_matrices is []:
            print("Decorators need to be created first! Call 'create_decorator_matrices()' first")
        return self.decorator_matrices


    '''
       Returns the list of previously created marker transformation matrices.
    '''
    def get_marker_matrices(self):
        if self.marker_matrices is []:
            print("Markers need to be created first! Call 'create_marker_matrices()' first")
        return self.marker_matrices


    '''
       Returns a list of transformation matrices representing the route's start,
       first corner, second corner and goal.
    '''
    def create_marker_matrices(self, l1, l2, l3, turn_mode):

        # define rotation matrix that turns by 90 deg (left)
        rot_mat_left = t3d.rotation_matrix(math.radians(90), [0.0, 1.0, 0.0])
        rot_mat_left[numpy.abs(rot_mat_left) < 0.0000001] = 0

        # define rotation matrix that turns by -90 deg (right)
        rot_mat_right = t3d.rotation_matrix(math.radians(-90), [0.0, 1.0, 0.0])
        rot_mat_right[numpy.abs(rot_mat_right) < 0.0000001] = 0

        self.marker_matrices = []

        # add start marker
        trans_mat_1 = t3d.translation_matrix([0.0, 0.0, 0.0])
        rot_mat_1 = t3d.rotation_matrix(math.radians(0), [0.0, 1.0, 0.0])
        rot_mat_1[numpy.abs(rot_mat_1) < 0.0000001] = 0
        self.marker_matrices.append(t3d.concatenate_matrices(trans_mat_1, rot_mat_1))

        # add corner 1 marker
        trans_mat_2 = t3d.translation_matrix([0.0, 0.0, -l1])
        if turn_mode is "LL":
            self.marker_matrices.append(t3d.concatenate_matrices(trans_mat_2, rot_mat_left))
        else:
            self.marker_matrices.append(t3d.concatenate_matrices(trans_mat_2, rot_mat_right))

        # add corner 2 marker
        trans_mat_3_1 = t3d.translation_matrix([0.0, 0.0, -l1])
        trans_mat_3_2 = t3d.translation_matrix([0.0, 0.0, -l2])
        if turn_mode is "LL":
            self.marker_matrices.append(
                t3d.concatenate_matrices(trans_mat_3_1, rot_mat_left, trans_mat_3_2, rot_mat_left))
        else:
            self.marker_matrices.append(
                t3d.concatenate_matrices(trans_mat_3_1, rot_mat_right, trans_mat_3_2, rot_mat_right))

        # add goal marker
        trans_mat_4_1 = t3d.translation_matrix([0.0, 0.0, -l1])
        trans_mat_4_2 = t3d.translation_matrix([0.0, 0.0, -l2])
        trans_mat_4_3 = t3d.translation_matrix([0.0, 0.0, -l3])
        if turn_mode is "LL":
            self.marker_matrices.append(
                t3d.concatenate_matrices(trans_mat_4_1, rot_mat_left, trans_mat_4_2, rot_mat_left, trans_mat_4_3))
        else:
            self.marker_matrices.append(
                t3d.concatenate_matrices(trans_mat_4_1, rot_mat_right, trans_mat_4_2, rot_mat_right, trans_mat_4_3))

        return self.marker_matrices
        

    '''
       Returns a list of transformation matrices for unit cubes to represent
       houses and decorator objects along the three-segment routes.
    '''
    def create_object_matrices(self, l1, l2, l3, turn_mode):

        self.house_matrices = []

        # scale segment parts
        lb2 = random.uniform(self.min_baseline_distance, self.max_baseline_distance)

        self.goal_extension = self.min_extension + random.uniform(l1, 1.5*l1)
        self.opposite_goal_extension = self.min_extension + random.uniform(l3, l3 + self.max_additional_extension)
        self.baseline_extension = random.uniform(0, self.path_width)

        self.build_roads(l1, l2, l3, lb2, turn_mode)


    '''
       Generates all the houses and decorators required for the city layout.
    '''
    def build_roads(self, l1, l2, l3, lb2, turn_mode):

        # define translation matrices
        trans_mat_l1 = t3d.translation_matrix([0.0, 0.0, -l1])
        trans_mat_l2 = t3d.translation_matrix([0.0, 0.0, -l2])

        # define rotation matrix that turns by 90 deg (left)
        rot_mat_left = t3d.rotation_matrix(math.radians(90), [0.0, 1.0, 0.0])
        rot_mat_left[numpy.abs(rot_mat_left) < 0.0000001] = 0

        # define rotation matrix that turns by -90 deg (right)
        rot_mat_right = t3d.rotation_matrix(math.radians(-90), [0.0, 1.0, 0.0])
        rot_mat_right[numpy.abs(rot_mat_right) < 0.0000001] = 0

        # define rotation matrix that turns by 180 deg (right)
        rot_mat_180 = t3d.rotation_matrix(math.radians(180), [0.0, 1.0, 0.0])
        rot_mat_180[numpy.abs(rot_mat_180) < 0.0000001] = 0


        # houses along l1
        self.more_start_extent_for_second_row = False

        remainder_1 = self.build_street_side_houses(base_matrix=numpy.identity(4),
                                                    start_free_room=-self.path_width,
                                                    first_house_max_depth=1000.0,
                                                    max_length=l1 - self.path_width / 2.0,
                                                    side='left',
                                                    generate_gaps=True)

        remainder_2 = self.build_street_side_houses(base_matrix=numpy.identity(4),
                                                    start_free_room=-self.path_width,
                                                    first_house_max_depth=1000.0,
                                                    max_length=l1 - self.path_width / 2.0,
                                                    side='right',
                                                    generate_gaps=True)

        # houses along l2 in all directions
        self.more_start_extent_for_second_row = True

        r3_base_mat = t3d.concatenate_matrices(trans_mat_l1, rot_mat_left)
        remainder_3 = self.build_street_side_houses(base_matrix=r3_base_mat,
                                                    start_free_room=self.path_width / 2.0,
                                                    first_house_max_depth=remainder_1,
                                                    max_length=l2 - self.path_width / 2.0,
                                                    side='left',
                                                    generate_gaps=True)

        r4_base_mat = t3d.concatenate_matrices(trans_mat_l1, rot_mat_right)
        remainder_4 = self.build_street_side_houses(base_matrix=r4_base_mat,
                                                    start_free_room=self.path_width / 2.0,
                                                    first_house_max_depth=remainder_2,
                                                    max_length=l2 - self.path_width / 2.0,
                                                    side='right',
                                                    generate_gaps=True)

        self.more_start_extent_for_second_row = False

        r5_base_mat = t3d.concatenate_matrices(trans_mat_l1, rot_mat_left, trans_mat_l2, rot_mat_180)
        remainder_5 = self.build_street_side_houses(base_matrix=r5_base_mat,
                                                    start_free_room=self.path_width / 2.0 + 10.0,
                                                    # allow space for house around corner
                                                    first_house_max_depth=1000.0,
                                                    max_length=2 * l2 - self.path_width / 2.0,
                                                    side='left',
                                                    generate_gaps=True)

        # houses along l3 in all directions
        if turn_mode == 'LL':
            self.more_start_extent_for_second_row = True

            base_mat = t3d.concatenate_matrices(trans_mat_l1, rot_mat_left, trans_mat_l2, rot_mat_left)
            self.build_street_side_houses(base_matrix=base_mat,
                                          start_free_room=self.path_width / 2.0,
                                          first_house_max_depth=remainder_3,
                                          max_length=l3 + self.goal_extension,
                                          side='left',
                                          generate_gaps=True)  # LL

            base_mat = t3d.concatenate_matrices(trans_mat_l1, rot_mat_left, trans_mat_l2, rot_mat_right)
            self.build_street_side_houses(base_matrix=base_mat,
                                          start_free_room=self.path_width / 2.0,
                                          first_house_max_depth=10.0,
                                          max_length=l3 + self.opposite_goal_extension,
                                          side='right',
                                          generate_gaps=True)  # LR

            self.more_start_extent_for_second_row = False

            trans_mat_3 = t3d.translation_matrix([0.0, 0.0, -l3 - self.goal_extension])
            base_mat = t3d.concatenate_matrices(trans_mat_l1, rot_mat_left, trans_mat_l2, rot_mat_left,
                                                     trans_mat_3, rot_mat_180)

            self.build_street_side_houses(base_matrix=base_mat,
                                          start_free_room=0.0,
                                          first_house_max_depth=1000.0,
                                          max_length=self.goal_extension + 2.0 * l3 + self.opposite_goal_extension,
                                          side='left',
                                          generate_gaps=True)

            trans_mat_3 = t3d.translation_matrix([0.0, 0.0, -l3 - self.goal_extension])
            base_mat = t3d.concatenate_matrices(trans_mat_l1, rot_mat_left, trans_mat_l2, rot_mat_left,
                                                     trans_mat_3)
            self.close_segment(base_matrix=base_mat)  # LL

            trans_mat_3 = t3d.translation_matrix([0.0, 0.0, -l3 - self.opposite_goal_extension])
            base_mat = t3d.concatenate_matrices(trans_mat_l1, rot_mat_left, trans_mat_l2, rot_mat_right,
                                                     trans_mat_3)
            self.close_segment(base_matrix=base_mat)  # LR

            trans_mat_2 = t3d.translation_matrix([0.0, 0.0, -l2 + self.path_width / 2.0])
            base_mat = t3d.concatenate_matrices(trans_mat_l1, rot_mat_right, trans_mat_2)
            self.close_segment(base_matrix=base_mat)

        if turn_mode == 'RR':
            self.more_start_extent_for_second_row = True

            base_mat = t3d.concatenate_matrices(trans_mat_l1, rot_mat_right, trans_mat_l2, rot_mat_left)
            self.build_street_side_houses(base_matrix=base_mat,
                                          start_free_room=self.path_width / 2.0,
                                          first_house_max_depth=remainder_5,
                                          max_length=l3 + self.opposite_goal_extension,
                                          side='left',
                                          generate_gaps=True)  # RL

            base_mat = t3d.concatenate_matrices(trans_mat_l1, rot_mat_right, trans_mat_l2, rot_mat_right)
            self.build_street_side_houses(base_matrix=base_mat,
                                          start_free_room=self.path_width / 2.0,
                                          first_house_max_depth=remainder_4,
                                          max_length=l3 + self.goal_extension,
                                          side='right',
                                          generate_gaps=True)  # RR

            self.more_start_extent_for_second_row = False

            trans_mat_3 = t3d.translation_matrix(
                [0.0, 0.0, -l3 - self.opposite_goal_extension])
            base_mat = t3d.concatenate_matrices(trans_mat_l1, rot_mat_right, trans_mat_l2, rot_mat_left,
                                                     trans_mat_3,
                                                     rot_mat_180)
            temp_length = self.opposite_goal_extension + 2 * l3 + self.goal_extension
            self.build_street_side_houses(base_matrix=base_mat, start_free_room=0.0,
                                          first_house_max_depth=1000.0,
                                          max_length=temp_length, side='left',
                                          generate_gaps=True)

            trans_mat_3 = t3d.translation_matrix([0.0, 0.0, -l3 - self.opposite_goal_extension])
            base_mat = t3d.concatenate_matrices(trans_mat_l1, rot_mat_right, trans_mat_l2, rot_mat_left,
                                                     trans_mat_3)
            self.close_segment(base_matrix=base_mat)  # RL

            trans_mat_3 = t3d.translation_matrix([0.0, 0.0, -l3 - self.goal_extension])
            base_mat = t3d.concatenate_matrices(trans_mat_l1, rot_mat_right, trans_mat_l2, rot_mat_right,
                                                     trans_mat_3)
            self.close_segment(base_matrix=base_mat)  # RR

            trans_mat_2 = t3d.translation_matrix([0.0, 0.0, -l2 + self.path_width / 2.0])
            base_mat = t3d.concatenate_matrices(trans_mat_l1, rot_mat_left, trans_mat_2)
            self.close_segment(base_matrix=base_mat)

        # path behind start (baseline task)
        self.more_start_extent_for_second_row = False
        trans_mat_1 = t3d.translation_matrix([0.0, 0.0, -self.path_width * 1.5])
        base_mat = t3d.concatenate_matrices(rot_mat_180, trans_mat_1, rot_mat_left)
        self.build_street_side_houses(base_matrix=base_mat,
                                      start_free_room=self.path_width / 2.0 + self.house_scale_vector[2],
                                      first_house_max_depth=1000.0,
                                      max_length=lb2 + self.baseline_extension,
                                      side='left',
                                      generate_gaps=True)

        trans_mat_1 = t3d.translation_matrix([0.0, 0.0, -self.path_width * 1.5])
        base_mat = t3d.concatenate_matrices(rot_mat_180, trans_mat_1, rot_mat_right)
        self.build_street_side_houses(base_matrix=base_mat,
                                      start_free_room=self.path_width / 2.0 + self.house_scale_vector[2],
                                      first_house_max_depth=1000.0,
                                      max_length=lb2 + self.baseline_extension,
                                      side='right',
                                      generate_gaps=True)

        trans_mat_1 = t3d.translation_matrix([0.0, 0.0, -self.path_width * 1.5])
        trans_mat_2 = t3d.translation_matrix([0.0, 0.0, -lb2-self.baseline_extension])
        base_mat = t3d.concatenate_matrices(rot_mat_180, trans_mat_1, rot_mat_left, trans_mat_2, rot_mat_180)
        self.build_street_side_houses(base_matrix=base_mat,
                                      start_free_room=0.0,
                                      first_house_max_depth=1000.0,
                                      max_length=2 * lb2 + 2 * self.baseline_extension,
                                      side='left',
                                      generate_gaps=True)

        trans_mat_1 = t3d.translation_matrix([0.0, 0.0, -self.path_width * 1.5])
        trans_mat_2 = t3d.translation_matrix([0.0, 0.0, -lb2-self.baseline_extension])
        base_mat = t3d.concatenate_matrices(rot_mat_180, trans_mat_1, rot_mat_left, trans_mat_2)
        self.close_segment(base_matrix=base_mat)

        trans_mat_1 = t3d.translation_matrix([0.0, 0.0, -self.path_width * 1.5])
        trans_mat_2 = t3d.translation_matrix([0.0, 0.0, -lb2-self.baseline_extension])
        base_mat = t3d.concatenate_matrices(rot_mat_180, trans_mat_1, rot_mat_right, trans_mat_2)
        self.close_segment(base_matrix=base_mat)


    '''
       Generates one row of houses and decorators along one side of a single street.
    '''
    def build_street_side_houses(self,
                                 base_matrix,
                                 start_free_room,
                                 first_house_max_depth,
                                 max_length,
                                 side,
                                 path_width=10.0,
                                 generate_gaps=False):
        current_length = start_free_room + self.house_scale_vector[0] / 2.0
        first_house = True
        side_sign = 1 if side == 'right' else -1

        # create second row with houses for view blinding
        if generate_gaps and self.enable_blinders:

            if self.more_start_extent_for_second_row:
                adjusted_start_free_room = start_free_room + 2*self.house_scale_vector[0] + self.path_width
            else:
                adjusted_start_free_room = start_free_room + self.house_scale_vector[0]

            self.build_street_side_houses(base_matrix=base_matrix,
                                          start_free_room=adjusted_start_free_room,
                                          first_house_max_depth=1000.0,
                                          max_length=max_length - self.house_scale_vector[0],
                                          side=side,
                                          path_width=path_width + self.house_scale_vector[2] * 4.0,
                                          generate_gaps = False)

        # place houses until termination criterion
        while True:

            if first_house:
                house_scl_vec = self.house_scale_vector
                z_scale = min(first_house_max_depth / self.house_scale_vector[2], 1.0)
                first_house = False
            else:
                house_scl_vec = self.house_scale_vector
                z_scale = 1.0

            extent = [a * b for a, b in zip([1.0, 1.0, 1.0], house_scl_vec)]

            # termination criterion 1
            if current_length + extent[0] / 2.0 > max_length:
                # if this case is active, no house will be placed at current_length anymore
                # return distance not covered by houses
                return (max_length - current_length) + extent[0] / 2.0  # last term since no house was placed

            # define rotation matrix that turns by 90 deg (left)
            rot_mat_90 = t3d.rotation_matrix(math.radians(side_sign * 90), [0.0, 1.0, 0.0])
            rot_mat_90[numpy.abs(rot_mat_90) < 0.0000001] = 0

            # define rotation matrix that turns by 180 deg (right)
            rot_mat_180 = t3d.rotation_matrix(math.radians(180), [0.0, 1.0, 0.0])
            rot_mat_180[numpy.abs(rot_mat_180) < 0.0000001] = 0

            trans_mat = t3d.translation_matrix(
                [side_sign * current_length, extent[1] / 2.0, path_width / 2.0 + extent[2] / 2.0])

            house_mat = t3d.concatenate_matrices(base_matrix, rot_mat_90, trans_mat, rot_mat_180,
                                                      self.house_scale_matrix)

            if z_scale != 1.0:
                adjust_house_scale_vector = [1.0, 1.0, z_scale]
                scl_mat_x = t3d.scale_matrix(adjust_house_scale_vector[0], origin=None, direction=[1, 0, 0])
                scl_mat_y = t3d.scale_matrix(adjust_house_scale_vector[1], origin=None, direction=[0, 1, 0])
                scl_mat_z = t3d.scale_matrix(adjust_house_scale_vector[2], origin=None, direction=[0, 0, 1])
                adjust_house_scale_mat = scl_mat_x * scl_mat_y * scl_mat_z

                adjust_position_mat_1 = t3d.translation_matrix(
                    [0.0, 0.0, extent[2] / 2.0])

                adjust_position_mat_2 = t3d.translation_matrix(
                    [0.0, 0.0, -extent[2] / 2.0])

                house_mat = t3d.concatenate_matrices(base_matrix, rot_mat_90, trans_mat, rot_mat_180,
                                                     adjust_position_mat_1, adjust_house_scale_mat,
                                                     adjust_position_mat_2, self.house_scale_matrix)

            self.house_matrices.append(house_mat)

            # decide if some decorator objects should be placed in front of the house
            if random.random() > (1.0-self.decorator_probability):
                decorator_trans_mat = t3d.translation_matrix([side_sign * current_length, 
                                                              0.0, self.path_width * 0.4])
                local_displacement = t3d.translation_matrix([random.uniform(-0.5, 0.5), 0.5, 0.0])
                decorator_mat = t3d.concatenate_matrices(base_matrix, rot_mat_90, decorator_trans_mat, 
                                                         local_displacement)
                self.decorator_matrices.append(decorator_mat)

            # generate gaps between houses if activated
            gap_modifier = 0.0
            no_gap_zone = (start_free_room + self.house_scale_vector[0] / 2.0) + self.house_scale_vector[0] * 2.0

            if generate_gaps and current_length > no_gap_zone and current_length < (max_length - no_gap_zone):
                random_number = random.random()
                if random_number > 0.8:
                    gap_modifier += self.path_width
                else:
                    gap_modifier += random.uniform(0.0, extent[0] * 0.5)

            next_current_length = current_length + extent[0] + gap_modifier

            # termination criterion 2
            if next_current_length > max_length:
                # if this case is active, a house was placed at current_length, but we will not go on
                # return distance not covered by houses
                return (max_length - current_length) - extent[0] / 2.0  # last term since house was placed

            current_length = next_current_length


    '''
       Closes the end of a street with houses.
    '''
    def close_segment(self, base_matrix, offset=0.0):
        current_width = -self.path_width / 2.0 + offset

        while True:
            if current_width >= self.path_width - offset:
                return

            trans_mat = t3d.translation_matrix([current_width, self.house_scale_vector[1] / 2.0, -self.house_scale_vector[2] / 2.0])
            house_mat = t3d.concatenate_matrices(base_matrix, trans_mat, self.house_scale_matrix)

            house_position = house_mat

            self.house_matrices.append(house_position)
            current_width += self.house_scale_vector[0]
